"""Fiona CLI commands."""

from collections import defaultdict
from copy import copy
import itertools
import json
import logging
import warnings

import click
from cligj import use_rs_opt  # type: ignore

from fiona.features import map_feature, reduce_features
from fiona.fio import with_context_env
from fiona.fio.helpers import obj_gen, eval_feature_expression  # type: ignore

log = logging.getLogger(__name__)


@click.command(
    "map",
    short_help="Map a pipeline expression over GeoJSON features.",
)
@click.argument("pipeline")
@click.option(
    "--raw",
    "-r",
    is_flag=True,
    default=False,
    help="Print raw result, do not wrap in a GeoJSON Feature.",
)
@click.option(
    "--no-input",
    "-n",
    is_flag=True,
    default=False,
    help="Do not read input from stream.",
)
@click.option(
    "--dump-parts",
    is_flag=True,
    default=False,
    help="Dump parts of geometries to create new inputs before evaluating pipeline.",
)
@use_rs_opt
def map_cmd(pipeline, raw, no_input, dump_parts, use_rs):
    """Map a pipeline expression over GeoJSON features.

    Given a sequence of GeoJSON features (RS-delimited or not) on stdin
    this prints copies with geometries that are transformed using a
    provided transformation pipeline. In "raw" output mode, this
    command prints pipeline results without wrapping them in a feature
    object.

    The pipeline is a string that, when evaluated by fio-map, produces
    a new geometry object. The pipeline consists of expressions in the
    form of parenthesized lists that may contain other expressions.
    The first item in a list is the name of a function or method, or an
    expression that evaluates to a function. The second item is the
    function's first argument or the object to which the method is
    bound. The remaining list items are the positional and keyword
    arguments for the named function or method. The names of the input
    feature and its geometry in the context of these expressions are
    "f" and "g".

    For example, this pipeline expression

        '(simplify (buffer g 100.0) 5.0)'

    buffers input geometries and then simplifies them so that no
    vertices are closer than 5 units. Keyword arguments for the shapely
    methods are supported. A keyword argument is preceded by ':' and
    followed immediately by its value. For example:

        '(simplify g 5.0 :preserve_topology true)'

    and

        '(buffer g 100.0 :resolution 8 :join_style 1)'

    Numerical and string arguments may be replaced by expressions. The
    buffer distance could be a function of a geometry's area.

        '(buffer g (/ (area g) 100.0))'

    """
    if no_input:
        features = [None]
    else:
        stdin = click.get_text_stream("stdin")
        features = obj_gen(stdin)

    for feat in features:
        for i, value in enumerate(map_feature(pipeline, feat, dump_parts=dump_parts)):
            if use_rs:
                click.echo("\x1e", nl=False)
            if raw:
                click.echo(json.dumps(value))
            else:
                new_feat = copy(feat)
                new_feat["id"] = f"{feat.get('id', '0')}:{i}"
                new_feat["geometry"] = value
                click.echo(json.dumps(new_feat))


@click.command(
    "filter",
    short_help="Evaluate pipeline expressions to filter GeoJSON features.",
)
@click.argument("pipeline")
@use_rs_opt
@click.option(
    "--snuggs-only",
    "-s",
    is_flag=True,
    default=False,
    help="Strictly require snuggs style expressions and skip check for type of expression.",
)
@click.pass_context
@with_context_env
def filter_cmd(ctx, pipeline, use_rs, snuggs_only):
    """Evaluate pipeline expressions to filter GeoJSON features.

    The pipeline is a string that, when evaluated, gives a new value for
    each input feature. If the value evaluates to True, the feature
    passes through the filter. Otherwise the feature does not pass.

    The pipeline consists of expressions in the form of parenthesized
    lists that may contain other expressions. The first item in a list
    is the name of a function or method, or an expression that evaluates
    to a function. The second item is the function's first argument or
    the object to which the method is bound. The remaining list items
    are the positional and keyword arguments for the named function or
    method. The names of the input feature and its geometry in the
    context of these expressions are "f" and "g".

    For example, this pipeline expression

        '(< (distance g (Point 4 43)) 1)'

    lets through all features that are less than one unit from the given
    point and filters out all other features.

    *New in version 1.10*: these parenthesized list expressions.

    The older style Python expressions like

        'f.properties.area > 1000.0'

    are deprecated and will not be supported in version 2.0.

    """
    stdin = click.get_text_stream("stdin")
    features = obj_gen(stdin)

    if not snuggs_only:
        try:
            from pyparsing.exceptions import ParseException
            from fiona._vendor.snuggs import ExpressionError, expr

            if not pipeline.startswith("("):
                test_string = f"({pipeline})"
            expr.parseString(test_string)
        except ExpressionError:
            # It's a snuggs expression.
            log.info("Detected a snuggs expression.")
            pass
        except ParseException:
            # It's likely an old-style Python expression.
            log.info("Detected a legacy Python expression.")
            warnings.warn(
                "This style of filter expression is deprecated. "
                "Version 2.0 will only support the new parenthesized list expressions.",
                FutureWarning,
            )
            for i, obj in enumerate(features):
                feats = obj.get("features") or [obj]
                for j, feat in enumerate(feats):
                    if not eval_feature_expression(feat, pipeline):
                        continue
                    if use_rs:
                        click.echo("\x1e", nl=False)
                    click.echo(json.dumps(feat))
            return

    for feat in features:
        for value in map_feature(pipeline, feat):
            if value:
                if use_rs:
                    click.echo("\x1e", nl=False)
                click.echo(json.dumps(feat))


@click.command("reduce", short_help="Reduce a stream of GeoJSON features to one value.")
@click.argument("pipeline")
@click.option(
    "--raw",
    "-r",
    is_flag=True,
    default=False,
    help="Print raw result, do not wrap in a GeoJSON Feature.",
)
@use_rs_opt
@click.option(
    "--zip-properties",
    is_flag=True,
    default=False,
    help="Zip the items of input feature properties together for output.",
)
def reduce_cmd(pipeline, raw, use_rs, zip_properties):
    """Reduce a stream of GeoJSON features to one value.

    Given a sequence of GeoJSON features (RS-delimited or not) on stdin
    this prints a single value using a provided transformation pipeline.

    The pipeline is a string that, when evaluated, produces
    a new geometry object. The pipeline consists of expressions in the
    form of parenthesized lists that may contain other expressions.
    The first item in a list is the name of a function or method, or an
    expression that evaluates to a function. The second item is the
    function's first argument or the object to which the method is
    bound. The remaining list items are the positional and keyword
    arguments for the named function or method. The set of geometries
    of the input features in the context of these expressions is named
    "c".

    For example, the pipeline expression

        '(unary_union c)'

    dissolves the geometries of input features.

    To keep the properties of input features while reducing them to a
    single feature, use the --zip-properties flag. The properties of the
    input features will surface in the output feature as lists
    containing the input values.

    """
    stdin = click.get_text_stream("stdin")
    features = (feat for feat in obj_gen(stdin))

    if zip_properties:
        prop_features, geom_features = itertools.tee(features)
        properties = defaultdict(list)
        for feat in prop_features:
            for key, val in feat["properties"].items():
                properties[key].append(val)
    else:
        geom_features = features
        properties = {}

    for result in reduce_features(pipeline, geom_features):
        if use_rs:
            click.echo("\x1e", nl=False)
        if raw:
            click.echo(json.dumps(result))
        else:
            click.echo(
                json.dumps(
                    {
                        "type": "Feature",
                        "properties": properties,
                        "geometry": result,
                        "id": "0",
                    }
                )
            )



"""Dataset paths, identifiers, and filenames

Note well: this module is deprecated in 1.3.0 and will be removed in a
future version.
"""

import warnings

from fiona._path import _ParsedPath as ParsedPath
from fiona._path import _UnparsedPath as UnparsedPath
from fiona._path import _parse_path as parse_path
from fiona._path import _vsi_path as vsi_path
from fiona.errors import FionaDeprecationWarning

warnings.warn(
    "fiona.path will be removed in version 2.0.", FionaDeprecationWarning
)



"""Operations on GeoJSON feature and geometry objects."""

from collections import UserDict
from functools import wraps
import itertools
from typing import Generator, Iterable, Mapping, Union

from fiona.transform import transform_geom  # type: ignore
import shapely  # type: ignore
import shapely.ops  # type: ignore
from shapely.geometry import mapping, shape  # type: ignore
from shapely.geometry.base import BaseGeometry, BaseMultipartGeometry  # type: ignore

from .errors import ReduceError
from ._vendor import snuggs

# Patch snuggs's func_map, extending it with Python builtins, geometry
# methods and attributes, and functions exported in the shapely module
# (such as set_precision).


class FuncMapper(UserDict, Mapping):
    """Resolves functions from names in pipeline expressions."""

    def __getitem__(self, key):
        """Get a function by its name."""
        if key in self.data:
            return self.data[key]
        elif key in __builtins__ and not key.startswith("__"):
            return __builtins__[key]
        elif key in dir(shapely):
            return lambda g, *args, **kwargs: getattr(shapely, key)(g, *args, **kwargs)
        elif key in dir(shapely.ops):
            return lambda g, *args, **kwargs: getattr(shapely.ops, key)(
                g, *args, **kwargs
            )
        else:
            return (
                lambda g, *args, **kwargs: getattr(g, key)(*args, **kwargs)
                if callable(getattr(g, key))
                else getattr(g, key)
            )


def collect(geoms: Iterable) -> object:
    """Turn a sequence of geometries into a single GeometryCollection.

    Parameters
    ----------
    geoms : Iterable
        A sequence of geometry objects.

    Returns
    -------
    Geometry

    """
    return shapely.GeometryCollection(list(geoms))


def dump(geom: Union[BaseGeometry, BaseMultipartGeometry]) -> Generator:
    """Get the individual parts of a geometry object.

    If the given geometry object has a single part, e.g., is an
    instance of LineString, Point, or Polygon, this function yields a
    single result, the geometry itself.

    Parameters
    ----------
    geom : a shapely geometry object.

    Yields
    ------
    A shapely geometry object.

    """
    if hasattr(geom, "geoms"):
        parts = geom.geoms
    else:
        parts = [geom]
    for part in parts:
        yield part


def identity(obj: object) -> object:
    """Get back the given argument.

    To help in making expression lists, where the first item must be a
    callable object.

    Parameters
    ----------
    obj : objeect

    Returns
    -------
    obj

    """
    return obj


def vertex_count(obj: object) -> int:
    """Count the vertices of a GeoJSON-like geometry object.

    Parameters
    ----------
    obj: object
        A GeoJSON-like mapping or an object that provides
        __geo_interface__.

    Returns
    -------
    int

    """
    shp = shape(obj)
    if hasattr(shp, "geoms"):
        return sum(vertex_count(part) for part in shp.geoms)
    elif hasattr(shp, "exterior"):
        return vertex_count(shp.exterior) + sum(
            vertex_count(ring) for ring in shp.interiors
        )
    else:
        return len(shp.coords)


def binary_projectable_property_wrapper(func):
    """Project func's geometry args before computing a property.

    Parameters
    ----------
    func : callable
        Signature is func(geom1, geom2, *args, **kwargs)

    Returns
    -------
    callable
        Signature is func(geom1, geom2, projected=True, *args, **kwargs)

    """

    @wraps(func)
    def wrapper(geom1, geom2, *args, projected=True, **kwargs):
        if projected:
            geom1 = shape(transform_geom("OGC:CRS84", "EPSG:6933", mapping(geom1)))
            geom2 = shape(transform_geom("OGC:CRS84", "EPSG:6933", mapping(geom2)))

        return func(geom1, geom2, *args, **kwargs)

    return wrapper


def unary_projectable_property_wrapper(func):
    """Project func's geometry arg before computing a property.

    Parameters
    ----------
    func : callable
        Signature is func(geom1, *args, **kwargs)

    Returns
    -------
    callable
        Signature is func(geom1, projected=True, *args, **kwargs)

    """

    @wraps(func)
    def wrapper(geom, *args, projected=True, **kwargs):
        if projected:
            geom = shape(transform_geom("OGC:CRS84", "EPSG:6933", mapping(geom)))

        return func(geom, *args, **kwargs)

    return wrapper


def unary_projectable_constructive_wrapper(func):
    """Project func's geometry arg before constructing a new geometry.

    Parameters
    ----------
    func : callable
        Signature is func(geom1, *args, **kwargs)

    Returns
    -------
    callable
        Signature is func(geom1, projected=True, *args, **kwargs)

    """

    @wraps(func)
    def wrapper(geom, *args, projected=True, **kwargs):
        if projected:
            geom = shape(transform_geom("OGC:CRS84", "EPSG:6933", mapping(geom)))
            product = func(geom, *args, **kwargs)
            return shape(transform_geom("EPSG:6933", "OGC:CRS84", mapping(product)))
        else:
            return func(geom, *args, **kwargs)

    return wrapper


area = unary_projectable_property_wrapper(shapely.area)
buffer = unary_projectable_constructive_wrapper(shapely.buffer)
distance = binary_projectable_property_wrapper(shapely.distance)
set_precision = unary_projectable_constructive_wrapper(shapely.set_precision)
simplify = unary_projectable_constructive_wrapper(shapely.simplify)
length = unary_projectable_property_wrapper(shapely.length)

snuggs.func_map = FuncMapper(
    area=area,
    buffer=buffer,
    collect=collect,
    distance=distance,
    dump=dump,
    identity=identity,
    length=length,
    simplify=simplify,
    set_precision=set_precision,
    vertex_count=vertex_count,
    **{
        k: getattr(itertools, k)
        for k in dir(itertools)
        if not k.startswith("_") and callable(getattr(itertools, k))
    },
)


def map_feature(
    expression: str, feature: Mapping, dump_parts: bool = False
) -> Generator:
    """Map a pipeline expression to a feature.

    Yields one or more values.

    Parameters
    ----------
    expression : str
        A snuggs expression. The outermost parentheses are optional.
    feature : dict
        A Fiona feature object.
    dump_parts : bool, optional (default: False)
        If True, the parts of the feature's geometry are turned into
        new features.

    Yields
    ------
    object

    """
    if not (expression.startswith("(") and expression.endswith(")")):
        expression = f"({expression})"

    try:
        geom = shape(feature.get("geometry", None))
        if dump_parts and hasattr(geom, "geoms"):
            parts = geom.geoms
        else:
            parts = [geom]
    except (AttributeError, KeyError):
        parts = [None]

    for part in parts:
        result = snuggs.eval(expression, g=part, f=feature)
        if isinstance(result, (str, float, int, Mapping)):
            yield result
        elif isinstance(result, (BaseGeometry, BaseMultipartGeometry)):
            yield mapping(result)
        else:
            try:
                for item in result:
                    if isinstance(item, (BaseGeometry, BaseMultipartGeometry)):
                        item = mapping(item)
                    yield item
            except TypeError:
                yield result


def reduce_features(expression: str, features: Iterable[Mapping]) -> Generator:
    """Reduce a collection of features to a single value.

    The pipeline is a string that, when evaluated by snuggs, produces
    a new value. The name of the input feature collection in the
    context of the pipeline is "c".

    Parameters
    ----------
    pipeline : str
        Geometry operation pipeline such as "(unary_union c)".
    features : iterable
        A sequence of Fiona feature objects.

    Yields
    ------
    object

    Raises
    ------
    ReduceError

    """
    if not (expression.startswith("(") and expression.endswith(")")):
        expression = f"({expression})"

    collection = (shape(feat["geometry"]) for feat in features)
    result = snuggs.eval(expression, c=collection)

    if isinstance(result, (str, float, int, Mapping)):
        yield result
    elif isinstance(result, (BaseGeometry, BaseMultipartGeometry)):
        yield mapping(result)
    else:
        raise ReduceError("Expression failed to reduce to a single value.")
